#!/usr/bin/python3

import os
import sys
from collections import OrderedDict, defaultdict
from ipaddress import ip_address as ip
import statistics
from typing import Dict, List

import dpkt
import matplotlib.ticker as mtick
import matplotlib.pyplot as plt
import math
import numpy as np

import process_generator

def autolabel(rects, labels, ax, xpos='center', format_str='{0:.0%}'):
    """
    Attach a text label above each bar in *rects*, displaying its height.

    *xpos* indicates which side to place the text w.r.t. the center of
    the bar. It can be one of the following {'center', 'right', 'left'}.
    """

    xpos = xpos.lower()  # normalize the case of the parameter
    ha = {'center': 'center', 'right': 'left', 'left': 'right'}
    offset = {'center': 0.5, 'right': 0.57, 'left': 0.43}  # x_txt = x + w*off

    for rect, label in zip(rects, labels):
        height = rect.get_height()
        ax.text(rect.get_x() + rect.get_width()*offset[xpos], 1.01*height,
                format_str.format(label), ha=ha[xpos], va='bottom')

def split_netem_string(netem):
    l = netem.split()
    if not l:
        return ""
    for w in ["delay", "loss", "random"]:
        if w in l:
            l.remove(w)

    if len(l) == 3:
        l[0] = l[0][:l[0].index(".")+2]+" ms"
        l[1] = l[1][:l[1].index(".")+2]+" ms"
        l[2] = l[2][:l[2].index(".")+2]+" %"
        return l[0]+" ±"+l[1]+"\n"+l[2]

    return "\n".join(l)



def plot_log_percentile_histogram(
            data: Dict[str, List[float]],
            title: str,
            config=None,
            outfile=None
        ) -> None:
    """
    For graph explanation, see
    https://blog.powerdns.com/2017/11/02/dns-performance-metrics-the-logarithmic-percentile-histogram/
    """
    fig, ax = plt.subplots(figsize=(16, 9))


    # Distribute sample points along logarithmic X axis
    percentiles = np.logspace(-3, 2, num=10000)

    def tick_formatter(x, _):
        packets = int(len(data[next(data.keys().__iter__())])/100*x)
        noun = "packet" + ("" if packets == 1 else "s")
        return "%s\n%d %s" % (x, packets, noun)

    ax.set_xscale('log')
    ax.xaxis.set_major_formatter(mtick.FuncFormatter(tick_formatter))
    ax.set_yscale('log')
    ax.yaxis.set_major_formatter(mtick.FormatStrFormatter('%s'))

    ax.grid(True, which='major')
    ax.grid(True, which='minor', linestyle='dotted', color='#DDDDDD')
    print(title)
    ax.set_xlabel('Slowest percentile')
    ax.set_ylabel('Response time [ms]')
    # ax.set_title('Resolver Response Time' + " - " + title)

    # plot data
    for server in sorted(data):
        if data[server]:
            try:
                color = config[server]['graph_color']
            except (KeyError, TypeError):
                color = None

            noerror = servers[server]["analysis"].stats["rcode"][0]
            servfail = servers[server]["analysis"].stats["rcode"][2]
            print(noerror)
            # convert to ms and sort
            values = sorted([1000 * x for x in data[server]], reverse=True)
            ax.plot(percentiles,
                    [values[math.ceil(pctl * len(values) / 100) - 1] for pctl in percentiles], lw=2,
                    label='{:<10} {:.0%} NOERROR, {} SERVFAIL'.format(server, noerror/len(values), servfail),
                    color=color)

    ax.legend(fontsize='large')
    if outfile:
        fig.savefig(outfile)



class PCAPAnalyzer:

    def __init__(self, pcap_path: str, scenario_dir: str):
        self.file = open(pcap_path, "rb")
        self.pcap_reader = dpkt.pcap.Reader(self.file)
        self.stats = defaultdict(lambda: defaultdict(int))
        self.scenario = scenario_dir
        self.config = process_generator.load_configs(self.scenario)
        self.resolver_ips = [ip(a) for a in self.config["resolver"]["address"]]
        self.client_ip = ip(self.config["client"]["address"][0])
        self.rtt_windows = []

    def __del__(self):
        self.file.close()

    def process_file(self):
        for ts, packet in self.pcap_reader:
            self.process_packet(ts, packet)
        q = self.stats["queries"]["from_client"]
        a = self.stats["answers"]["from_resolver"]
        if q != a:
            raise Warning("Resolver didn't answer all questions from client. %d queries and %d answers." % (q, a))

    def process_packet(self, ts, packet):
        for member_name in dir(self):
            if member_name.startswith("_packet"):
                # Unpack Linux cooked capture
                l2_packet = dpkt.sll.SLL(packet)
                getattr(self, member_name)(ts, l2_packet.data)

    def _packet_count_per_auth_ip(self, ts, ip_packet):
        ip_src = ip(ip_packet.src)
        ip_dst = ip(ip_packet.dst)
        if ip_src in self.resolver_ips and ip_dst != self.client_ip:
            self.stats["ip_dst"][ip_dst] += 1

    def _packet_count_per_server(self, ts, ip_packet):
        ip_src = ip(ip_packet.src)
        ip_dst = ip(ip_packet.dst)
        if ip_src in self.resolver_ips and ip_dst != self.client_ip:
            auth = self.identify_server(ts, ip_dst)
            if auth is None:
                auth = {}
                auth["name"] = "unknown("+str(ip_dst)+")"
                self.stats["packets"]["to_nonexistent_server"] += 1
            self.stats["server_dst"][auth["name"]] += 1
            self.stats["auth_config"][auth["name"]] = auth

    def _packet_dns_answer_rcode_count(self, ts, ip_packet):
        ip_src = ip(ip_packet.src)
        ip_dst = ip(ip_packet.dst)
        if ip_src in self.resolver_ips and ip_dst == self.client_ip:
            dns_packet = dpkt.dns.DNS(ip_packet.data.data)
            self.stats["rcode"][dns_packet.rcode] += 1

    def _packet_dns_query_count(self, ts, ip_packet):
        ip_src = ip(ip_packet.src)
        ip_dst = ip(ip_packet.dst)
        if ip_src in self.resolver_ips:
            if ip_dst != self.client_ip:
                self.stats["queries"]["from_resolver"] += 1
            else:
                self.stats["answers"]["from_resolver"] += 1
        if ip_dst in self.resolver_ips:
            if ip_src == self.client_ip:
                self.stats["queries"]["from_client"] += 1
            else:
                self.stats["answers"]["from_auths"] += 1


    def _packet_get_rtts(self, ts, ip_packet):
        ip_src = ip(ip_packet.src)
        ip_dst = ip(ip_packet.dst)
        if ip_src == self.client_ip:
            dns_packet = dpkt.dns.DNS(ip_packet.data.data)
            self.stats["rtts"][dns_packet.id] = -ts
        if ip_src in self.resolver_ips and ip_dst == self.client_ip:
            dns_packet = dpkt.dns.DNS(ip_packet.data.data)
            self.stats["rtts"][dns_packet.id] += ts

    def sliding_window_rtt(self, name, window = 100):

        fig, ax = plt.subplots(figsize=(16, 9))

        rtts = list(self.stats["rtts"].values())
        for i in range(window, len(rtts)):
            self.rtt_windows.append(statistics.mean(rtts[i-window:i]))


    def identify_server(self, ts, ip_addr):
        # This is slow :/
        candidates = []
        for path, config in self.config.items():
            if path.startswith("auth"):
                addresses = [ip(addr) for addr in config["address"]]
                if ip_addr in addresses:
                    candidates.append(config)

        if len(candidates) == 1:
            return candidates[0]

        for config in candidates:
            if config["starttime"] <= ts <= config["stoptime"]:
                return config

    def plot_stat(self, stat: str, resolver_name: str):
        fig, ax = plt.subplots(figsize=(16, 9))
        d = self.stats[stat]
        d = OrderedDict(sorted(d.items(), key=lambda t: t[1], reverse=True))
        print(d)
        ax.bar(range(len(d)), list(d.values()), align='center')
        plt.xticks(range(len(d)), list(d.keys()), rotation='vertical')
        # ax.set_title("%s - %s - %s" % (os.path.basename(self.scenario), stat, resolver_name))
        plt.tight_layout()
        plt.savefig(sys.argv[1] + "/" + resolver_name + "_"+stat+".svg", bbox_inches='tight', pad_inches=0.1)

    def plot_packet_distribution(self, name):
        fig, ax = plt.subplots(figsize=(16, 9))
        d = self.stats["server_dst"]
        d = OrderedDict(sorted(d.items(), key=lambda t: t[1], reverse=True))
        print(d)
        rects = ax.bar(range(len(d)), list(d.values()), align='center')
        config = self.stats["auth_config"]
        ticks = [key + "\n" + split_netem_string(config[key].get("netem", "")) for key in d.keys()]
        plt.xticks(range(len(d)), ticks, rotation='vertical', fontsize=12)
        # ax.set_title("%s - %s - %s" % (os.path.basename(self.scenario), "server_dst", name))

        q = self.stats["queries"]["from_resolver"]
        percentages = [k/q for k in d.values()]
        autolabel(rects, percentages, ax)

        plt.tight_layout()
        plt.savefig(sys.argv[1] + "/" + name + "_packet_distribution.svg", pad_inches=0.1, orientation='landscape', papertype="a4")


def plot_resolver_queries(server: Dict, scenario):
    data = {}
    percentages = {}
    q = 0
    for name in servers:
        q = servers[name]["analysis"].stats["queries"]["from_client"]
        data[name] = servers[name]["analysis"].stats["queries"]["from_resolver"]
        percentages[name] = data[name]/q
    fig, ax = plt.subplots(figsize=(16, 9))
    rects = ax.bar(range(len(data)), list(data.values()), align='center')
    plt.xticks(range(len(data)), list(data.keys()), rotation='vertical')
    # ax.set_title("%s - %s" % (os.path.basename(scenario), "queries from resolver"))
    plt.axhline(y=q,linewidth=1, color='k', ls="--")
    plt.text(3.45, q+75, "100%")
    print(percentages)
    autolabel(rects, percentages.values(), ax)
    plt.tight_layout()
    plt.savefig(sys.argv[1]+"/resolver_queries.svg", bbox_inches='tight', pad_inches=0.1)

def plot_lost_packets(server: Dict, scenario):
    data = {}
    for name in servers:
        data[name] = servers[name]["analysis"].stats["packets"]["to_nonexistent_server"]
    plt.rcParams["font.family"] = "monospace"
    fig, ax = plt.subplots(figsize=(16, 9))
    ax.bar(range(len(data)), list(data.values()), align='center')
    plt.xticks(range(len(data)), list(data.keys()), rotation='vertical')
    # ax.set_title("%s - %s" % (os.path.basename(scenario), "queries to dead address"))

    plt.tight_layout()
    plt.savefig(sys.argv[1]+"/lost_packets.svg", bbox_inches='tight', pad_inches=0.1)

def plot_servfails(server: Dict, scenario):
    data = {}
    for name in servers:
        data[name] = servers[name]["analysis"].stats["rcode"][2]
    plt.rcParams["font.family"] = "monospace"
    fig, ax = plt.subplots(figsize=(16, 9))
    ax.bar(range(len(data)), list(data.values()), align='center')
    plt.xticks(range(len(data)), list(data.keys()), rotation='vertical')
    # ax.set_title("%s - %s" % (os.path.basename(scenario), "SERVFAILs out of 10000"))
    plt.tight_layout()
    plt.savefig(sys.argv[1]+"/servfails.svg", bbox_inches='tight', pad_inches=0.1)
    fig.show()

def plot_rtts(servers: Dict):
    data = {}
    config = {}

    a_server_name = None
    for name in servers:
        a_server_name = name
        data[name] = servers[name]["analysis"].stats["rtts"].values()
        config[name] = {}
        config[name]["graph_color"] = servers[name]["color"]

    outfile = os.path.join(servers[a_server_name]["analysis"].scenario, "histogram.svg")
    print(os.path.basename(servers[a_server_name]["analysis"].scenario))
    plot_log_percentile_histogram(data, os.path.basename(servers[a_server_name]["analysis"].scenario), config, outfile)


def boxplot_rtts(servers: Dict):
    plt.rcParams["font.family"] = "monospace"
    _, ax = plt.subplots(figsize=(16, 9))
    ax.boxplot([list(s["analysis"].stats["rtts"].values()) for s in servers.values()], whis=[1,99])
    ax.set_xticklabels(servers.keys())

def compare_sliding_windows(servers, window=100):
    for name in servers:
        servers[name]["analysis"].sliding_window_rtt(window)

    fig, ax = plt.subplots(figsize=(16, 9))
    ax.set_title("sliding window = %d queries" % window)

    plt.ylim(0, 2)

    for name in servers:
        ax.plot(servers[name]["analysis"].rtt_windows, label=name, color=servers[name]["color"])

    ax.legend()
    plt.show()

def mp_worker_analyze(server):
    server["analysis"].process_file()


if __name__ == "__main__":
    scenario = sys.argv[1]
    names = [fn[7:-5] for fn in os.listdir(scenario) if fn.endswith('.pcap') and fn.startswith('result_')]
    services = names
    colors = [None] * len(names)
    pcaps = [os.path.join(scenario, "result_%s.pcap" % service) for service in services]
    analyses = [PCAPAnalyzer(pcap, scenario) for pcap in pcaps]

    plt.rcParams["font.family"] = "Liberation Sans"
    plt.rcParams["font.size"] = "18"

    servers = {}
    for name, service, color, pcap, analysis in zip(names, services, colors, pcaps, analyses):
        servers[name] = dict(service=service, color=color, pcap=pcap, analysis=analysis)
        servers[name]["analysis"].process_file()

#    import multiprocessing.pool
#    with multiprocessing.pool.Pool() as workerpool:
#        workerpool.map(mp_worker_analyze, servers.values())

    plot_rtts(servers)
    plot_resolver_queries(servers, os.path.basename(scenario))
    plot_servfails(servers, scenario)
    #compare_sliding_windows(servers, 100)

    # for i, a in enumerate(analyses):
    #      a.plot_packet_distribution(names[i])





